/*                                                               RTAUDIO.C   */

/*  This module is included when RTAUDIO is defined at compile time.
    It provides an interface between Csound realtime record/play calls
    and the device-driver code that controls the actual hardware.
    Makefile defines must have ONE of:  -DDEC, -DSGI, -Dsun, -DNeXT, -DHP, -DLINUX
 */

#include "cs.h"
#include "soundio.h"

#ifdef DEC
#include "DECplay.h"   /* doesn't include record feature */
#endif

#ifdef SGI
#include <stdio.h>
#include <audio.h>
static  ALconfig iconfig, oconfig;
static  ALport   iport = NULL, oport = NULL;
#endif

#ifdef sun
#include <sun/audioio.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
static  audio_info_t  audinfo;
static  int     audiofd, sampcnt = 0;
#endif

#ifdef sol
#include <sys/audioio.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
static  audio_info_t  audinfo;
        int     audiofd, sampcnt = 0;
#endif

#ifdef HP
void AWopen(), AudioWrite(), AWclose();
#endif

#ifdef NeXT
#include <stdio.h>
#include <sound/sound.h>
#include <sound/sounddriver.h>
#include <mach.h>
int low_water = 48*1024;   /* Used by the driver to control the flow of samples */
int high_water = 64*1024;  /* Used by the driver to control the flow of samples */
static port_t dev_port, owner_port, write_port, reply_port ;
#endif

#ifdef BLASTER
void sbinit(float, int, int, int);
int  sbplay(char *, int);
void sbfinish(void);
#endif

#ifdef LINUX			              /* Jonathan Mohr  1995 Oct 17 */
#include <unistd.h>
#include <fcntl.h>
#define DSP_NAME    "/dev/dsp"
static int dspfd;
void setsndparms(int, int, int, float, unsigned);
void setvolume(volume);
#endif

static  int     ishift = 0, oshift = 0, oMaxLag;
extern  long    nrecs;
extern	OPARMS	O;

static int getshift(int dsize)	/* turn sample- or frame-size into shiftsize */
{
        switch(dsize) {
	case 1:  return(0);
	case 2:  return(1);
	case 4:  return(2);
	case 8:  return(3);
	default: die("rtaudio: illegal dsize");
	}
}

void recopen(int nchnls, int dsize, float esr, int scale)
				/* open for audio input */
{
	oMaxLag = O.oMaxLag;        /* import DAC setting from command line   */
	if (oMaxLag <= 0)           /* if DAC sampframes ndef in command line */
	    oMaxLag = IODACSAMPS;   /*    use the default value               */
#ifdef DEC
/*        rec_set(nchnls, dsize, esr, scale);  */
        die("DECaudio record not available");
#endif
#ifdef SGI
	iconfig = ALnewconfig();
	ALsetchannels(iconfig, (long)nchnls);
	ALsetwidth(iconfig, (long)dsize);
/*	ALsetsamplerate(iconfig, (long)esr);    */
	{   long cmdBuf[2];
	    cmdBuf[0] = AL_INPUT_RATE;
	    cmdBuf[1] = (long)esr;
	    ALsetparams(AL_DEFAULT_DEVICE,cmdBuf,2);
	}
	ALsetqueuesize(iconfig, (long)oMaxLag);
	iport = ALopenport("soundi", "r", iconfig);
#endif
#ifdef sun
	die("SUN audio record not available");
#endif
#ifdef HP
	die("HP audio record not available");
#endif
#ifdef NeXT
	die("NeXT audio record not available");
#endif
#ifdef BLASTER
	die("SoundBlaster audio record not available (yet?)");
#endif
#ifdef LINUX                                  /* Jonathan Mohr  1995 Oct 17 */
	/* open DSP device for reading */
	if ( (dspfd = open(DSP_NAME, O_RDONLY)) == -1 )
	  die("unable to open soundcard for audio input");

	/* initialize data format, channels, sample rate, and buffer size */
	setsndparms( dspfd, O.informat, nchnls, esr, oMaxLag * O.insampsiz );
#endif
	ishift = getshift(dsize);
}

void playopen(int nchnls, int dsize, float esr, int scale)
				/* open for audio output */
{
        int b;

	oMaxLag = O.oMaxLag;        /* import DAC setting from command line   */
	if (oMaxLag <= 0)           /* if DAC sampframes ndef in command line */
	    oMaxLag = IODACSAMPS;   /*    use the default value               */
#ifdef DEC
        if ((b = O.outbufsamps / nchnls) >= 8192) {
	    sprintf(errmsg, "-b %d probably too large, suggest <= 2048", b);
	    warning(errmsg);
	}
        if (oMaxLag * esr / 11025 > 2048) {
	    sprintf(errmsg, "-B %d probably too large, suggest 1024", oMaxLag);
	    warning(errmsg);
	}
        play_set(nchnls, dsize, esr, scale);
	oshift = getshift(nchnls * dsize);
#endif
#ifdef SGI
    {	int PMLqueuesize;
	oconfig = ALnewconfig();
	ALsetchannels(oconfig, (long)nchnls);
	ALsetwidth(oconfig, (long)dsize);
/*	ALsetsamplerate(oconfig, (long)esr);    */
	{   long cmdBuf[2];
	    cmdBuf[0] = AL_OUTPUT_RATE;
	    cmdBuf[1] = (long)esr;
	    ALsetparams(AL_DEFAULT_DEVICE,cmdBuf,2);
	}
	ALsetqueuesize(oconfig, (long)oMaxLag);
	PMLqueuesize = ALgetqueuesize(oconfig);
	printf("\n PMLqueuesize = %d\n", PMLqueuesize);
	oport = ALopenport("soundo", "w", oconfig);
	oshift = getshift(dsize);
    }
#endif
#ifdef sun
	if (oMaxLag < 1024) {
	    sprintf(errmsg, "-B %d probably too small, suggest 1024", oMaxLag);
	    warning(errmsg);
	}
	if ((audiofd = open("/dev/audio", O_WRONLY)) < 0)
	    die("sfinit: cannot open /dev/audio");
	oshift = getshift(nchnls * dsize);
#endif
#ifdef HP
	AWopen(nchnls, (long)esr, oMaxLag);
/*	if ((audiofd = open("/dev/audio", O_WRONLY)) < 0)
	    die("sfinit: cannot open /dev/audio");  */
	oshift = getshift(nchnls * dsize);
#endif
#ifdef NeXT
    {   int protocol, sampleRate;
	if (esr > 33075.0)
	    sampleRate = SNDDRIVER_STREAM_TO_SNDOUT_44 ;
	else sampleRate = SNDDRIVER_STREAM_TO_SNDOUT_22 ;
	if (esr != 44100.0 && esr != 22050) {
	    long isr = esr, actual = isr > 33075 ? 44100 : 22050;
	    sprintf(errmsg,"SRate 44.1K or 22.05K only. This %ld playing at %ld",
		    isr, actual);
	    warning(errmsg);
	}
	if (nchnls != 2)
	    die("NeXT supports nchnls = 2 (stereo) output only");
	SNDAcquire(SND_ACCESS_OUT,0,0,0,NULL_NEGOTIATION_FUN,0,&dev_port,&owner_port); 
	snddriver_stream_setup(dev_port, owner_port, 
			sampleRate,  /* make this a switch */
			4096, 2, low_water, high_water, &protocol, &write_port);
	snddriver_stream_control(write_port, 0, SNDDRIVER_PAUSE_STREAM);
/*	oshift = getshift(dsize);   */  /* what's correct here? oshift currently 0 */
    }                                   /* perhaps driver-writing takes bytes ?    */
#endif
#ifdef BLASTER
	err_printf( "Opening Blaster\n");
	switch (O.outformat) {
		default:
		case AE_ALAW:
		case AE_ULAW:
			sprintf(errmsg, "Forcing 8bit -c sound format\n");
			warning(errmsg);
		case AE_CHAR:
			O.outformat = AE_CHAR;
			break;
		case AE_LONG:
		case AE_FLOAT:
			sprintf(errmsg, "Forcing 16bit -s sound format\n");
			warning(errmsg);
		case AE_SHORT:
			O.outformat = AE_SHORT;
	}
	if (oMaxLag > 64*1024) {
	    sprintf(errmsg, "-B %d too large, maximum is 64K", oMaxLag);
	    warning(errmsg);
	    O.oMaxLag = 32*1024;
	}
	else if (oMaxLag < 1024) {
	    sprintf(errmsg, "-B %d possibly too small, suggest minimum is 1K", oMaxLag);
	    warning(errmsg);
	}
	sbinit(esr, oMaxLag, nchnls, O.outformat);
	err_printf( "It's open\n");
	oshift = getshift(dsize);
#endif
#ifdef LINUX			                   /*  J. Mohr  1995 Oct 17 */
/* The following code not only opens the DSP device (soundcard and driver)
   for writing and initializes it for the proper sample size, rate, and
   channels, but allows the user to set the output volume.                  */
    {
	/* open DSP device for writing */
	if ( (dspfd = open(DSP_NAME, O_WRONLY)) == -1 )
	  die("unable to open soundcard for audio output");

	/* set sample size/format, rate, channels, and DMA buffer size */
	setsndparms( dspfd, O.outformat, nchnls, esr,
		     O.outbufsamps * O.outsampsiz);

	/* check if volume was specified as command line parameter */
	if (O.Volume) {
	  /* check range of value specified */
	  if (O.Volume > 100 || O.Volume < 0)
	    die("Volume must be between 0 and 100");
	  setvolume( O.Volume );
	}

	/* 'oshift' is not currently used by the Linux driver, but ... */
	oshift = getshift(nchnls * dsize);
    }
#endif
}

int rtrecord(char *inbuf, int nbytes) /* get samples from ADC */
{
#ifdef DEC
/*      rec_on(inbuf, (long)nbytes >> ishift);   */  /* not yet supported */
#endif
#ifdef SGI
	ALreadsamps(iport, inbuf, (long)nbytes >> ishift);
#endif
#ifdef LINUX			                   /*  J. Mohr  1995 Oct 17 */
	if ( (nbytes = read(dspfd, inbuf, nbytes)) == -1 )
	  die("error while reading DSP device for audio input");
#endif
	return(nbytes);
}

void rtplay(char *outbuf, int nbytes) /* put samples to DAC  */
    /* N.B. This routine serves as a THROTTLE in Csound Realtime Performance, */
    /* delaying the actual writes and return until the hardware output buffer */
    /* passes a sample-specific THRESHOLD.  If the I/O BLOCKING functionality */
    /* is implemented ACCURATELY by the vendor-supplied audio-library write,  */
    /* that is sufficient.  Otherwise, requires some kind of IOCTL from here. */
    /* This functionality is IMPORTANT when other realtime I/O is occurring,  */
    /* such as when external MIDI data is being collected from a serial port. */
    /* Since Csound polls for MIDI input at the software synthesis K-rate     */
    /* (the resolution of all software-synthesized events), the user can      */
    /* eliminate MIDI jitter by requesting that both be made synchronous with */
    /* the above audio I/O blocks, i.e. by setting -b to some 1 or 2 K-prds.  */
{
        long sampframes = nbytes >> oshift;
#ifdef DEC
        while (get_playbuf_remains() > oMaxLag);  /* once lag is below threshold, */
        play_on(outbuf, sampframes);              /*    send these sample-frames  */
#endif
#ifdef SGI
	ALwritesamps(oport, outbuf, sampframes);
#endif
#ifdef sun
	do {
	    if (ioctl(audiofd, AUDIO_GETINFO, &audinfo) < 0)
	        die("/dev/audio: cannot do AUDIO_GETINFO");
	} while (sampcnt - audinfo.play.samples > oMaxLag);
	sampcnt += sampframes;
	if (write(audiofd, outbuf, nbytes) < nbytes)
	    printf("/dev/audio: couldn't write all bytes requested\n");
#endif
#ifdef HP
	AudioWrite(outbuf, nbytes);
	/*  should synchcronize with buffer writes here by a wait loop */
/*	do {
	    if (ioctl(audiofd, AUDIO_GETINFO, &audinfo) < 0)
	        die("/dev/audio: cannot do AUDIO_GETINFO");
	} while (sampcnt - audinfo.play.samples > oMaxLag);
*/
#endif
#ifdef NeXT
	snddriver_stream_start_writing(write_port,outbuf,
			sampframes,1,0,0,1,1,1,1,1,0, reply_port);	
	if(nrecs == 3)
	    snddriver_stream_control(write_port, 0, SNDDRIVER_RESUME_STREAM);
#endif
#ifdef BLASTER
	sbplay(outbuf, nbytes);
#endif
#ifdef LINUX			                   /*  J. Mohr  1995 Oct 17 */
	if (write(dspfd, outbuf, nbytes) < nbytes)
	    printf("/dev/dsp: couldn't write all bytes requested\n");
#endif
	nrecs++;
}

void rtclose(void)		/* close the I/O device entirely  */
{				/* called only when both complete */
#ifdef DEC
        play_rls();
#endif
#ifdef SGI
	if (iport != NULL)
	    ALcloseport(iport);
	if (oport != NULL) {
	    while (ALgetfilled(oport) > 0)
	        sginap(1);
	    ALcloseport(oport);
	}
#endif
#ifdef HP
	AWclose();
#endif
#ifdef NeXT
	if (dev_port) {
	    SNDRelease(SND_ACCESS_OUT,dev_port, owner_port) ;
	    port_deallocate(task_self(),reply_port);
	} 
#endif
#ifdef BLASTER
	sbfinish();
#endif
#ifdef LINUX			                   /*  J. Mohr  1995 Oct 17 */
	if (close(dspfd) == -1)
	  die("unable to close DSP device");
#endif
}

