/*      vdelay, multitap, reverb2 coded by Paris Smaragdis 1994 */
/*      Berklee College of Music Csound development team        */
/*      Copyright (c) May 1994.  All rights reserved            */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "cs.h"
#include "vdelay.h"
#include "revsets.h"

#define ESR     ((double)esr/1000.0)

void vdelset(VDEL *p)           /*  vdelay set-up   */
{
    unsigned long n;
    float *buf;

    if (p->aux.auxp == NULL)    /*      allocate space for delay buffer */
        auxalloc((long)(*p->imaxd * ESR * sizeof(float)), &p->aux);

    p->left = 0;

    buf = (float *)p->aux.auxp;  /*    make sure buffer is empty       */
    n = (unsigned long)*p->imaxd * ESR;
    do {
        *buf++ = 0;
    } while (--n);
}

void vdelay(VDEL *p)              /*      vdelay  routine */
{
    unsigned long  nn, maxd, indx;
    float *out = p->sr;  /* assign object data to local variables   */
    float *in = p->ain;
    float *del = p->adel;
    float *buf = (float *)p->aux.auxp;

    maxd = (unsigned long) (*p->imaxd * ESR);
    nn = ksmps;
    indx = p->left;

    if (p->XINCODE == 3)    {    /*      if delay is a-rate      */
        do {
            register float  fv1, fv2;
            register long   v1, v2;
        
            buf[indx] = *in++;
        
                        
            fv1 = indx - (*del++) * ESR;        /* Make sure Inside the buffer      */
            while (fv1 > maxd)
                fv1 -= maxd;
            while (fv1 < 0)
                fv1 += maxd;
        
            if (fv1 < maxd - 1) /* Find next sample for interpolation      */
                fv2 = fv1 + 1;
            else
                fv2 = 0;
        
            v1 = (long)fv1;
            v2 = (long)fv2;
            *out++ = buf[v1] + (fv1 - ( long)fv1) * ( buf[v2] - buf[v1]);
        
            indx++;             /* Advance current pointer */
            /*indx %= maxd;*/
            while (indx >= maxd)
                indx -= maxd;

        } while (--nn);
    }
    else {                      /* and, if delay is k-rate */
        do {
            register float  fv1, fv2;
            register long   v1, v2;
        
            buf[indx] = *in++;
        
            fv1 = indx - *del * ESR; /* Make sure inside the buffer      */
            while (fv1 > maxd)
                fv1 -= maxd;
            while (fv1 < 0)
                fv1 += maxd;
        
            if (fv1 < maxd - 1) /* Find next sample for interpolation      */
                fv2 = fv1 + 1;
            else
                fv2 = 0;
        
            v1 = (long)fv1;
            v2 = (long)fv2;
            *out++ = buf[v1] + (fv1 - ( long)fv1) * ( buf[v2] - buf[v1]);
        
            indx++;             /*      Advance current pointer */
            /*indx %= maxd;*/
            while (indx >= maxd)
                indx -= maxd;
                        
        } while (--nn);
    }
    p->left = indx;             /*      and keep track of where you are */
}

void multitap_set(MDEL *p)
{
    register long n;
    register float *buf, max;

    if (p->INOCOUNT/2 == (float)p->INOCOUNT/2.)
        die("Wrong input count in multitap\n");

    for (n = 0 ; n < p->INOCOUNT - 1 ; n += 2) {
        if (max < *p->ndel[n]) max = *p->ndel[n];
    }

    if (p->aux.auxp == NULL)    /* allocate space for delay buffer */
        auxalloc((long)(esr * max * sizeof(float)), &p->aux);

    p->left = 0;
    p->max = (long)(esr * max);

    buf = (float *)p->aux.auxp; /* make sure buffer is empty       */
    n = p->max;
    do {
        *buf++ = 0;
    } while (--n);
}

void multitap_play(MDEL *p)
{                               /* assign object data to local variables   */
    register long  n, nn = ksmps, indx = p->left, delay;
    register float *out = p->sr, *in = p->ain;
    register float *buf = (float *)p->aux.auxp;

    do{
        buf[indx] = *in++;      /*      Write input     */
        *out = 0;               /*      Clear output buffer     */

        indx++;         /*      Advance input pointer   */
        indx %= p->max;

        for (n = 0 ; n < p->INOCOUNT - 1 ; n += 2) {

            delay = indx - (long)(esr * *p->ndel[n]);
            if (delay < 0)
                delay += p->max;

            *out += buf[delay] * *p->ndel[n+1]; /*      Write output    */
        }
        out++;                  /*      Advance output pointer  */

    } while (--nn);
    p->left = indx;
}

#define LOG001  (-6.9078)       /* log(.001) */

void reverb2_set(STVB *p)            /* 6-comb/lowpass, 5-allpass reverberator */
{
    register long i, n;
    register float *temp;

    if (*p->hdif > 1.0 || *p->hdif < 0)
        die("High frequency diffusion not in (0, 1)\n");

    for (i = 0 ; i < Combs ; i++) {
        p->c_time[i] = gc_time[i];
        p->c_gain[i] = exp((double)(LOG001 * (p->c_time[i]/esr) /
                                    (gc_gain[i] * *p->time)));
        p->g[i] = *p->hdif;
        p->c_gain[i] = p->c_gain[i] * (1 - p->g[i]);
        p->z[i] = 0.0;

        auxalloc((long)(p->c_time[i] * sizeof(float)), &p->caux[i]);
        p->cbuf_cur[i] = (float *)p->caux[i].auxp;
        for (n = 0 ; n < p->c_time[i] ; n++)
            *(p->cbuf_cur[i] + n) = 0.0;
    }

    for (i = 0 ; i < Alpas ; i++) {
        p->a_time[i] = ga_time[i];
        p->a_gain[i] = exp((double)(LOG001 * (p->c_time[i]/esr) /
                                    (ga_gain[i] * *p->time)));
        auxalloc((long) p->a_time[i] * sizeof(float), &p->aaux[i]);
        p->abuf_cur[i] = (float *)p->aaux[i].auxp;
    }

    auxalloc(ksmps * sizeof(float), &p->temp);
    temp = (float *)p->temp.auxp;
    for (n = 0 ; n < ksmps ; n++)
        *temp++ = 0.0;

    p->prev_time = *p->time;
    p->prev_hdif = *p->hdif;
}

int reverb2_play(STVB *p)
{
    register long       i, n = ksmps;
    register float      *in, *out = p->out, *buf, *end;
             float      gain, z;

    do  *out++ = 0.0;
    while (--n);

    if (*p->time != p->prev_time || *p->hdif != p->prev_hdif) {
        if (*p->hdif > 1.) {
            printf("Warning: High frequency diffusion>1\n");
            *p->hdif = 1.;
        }
        if (*p->hdif < 0)       {
            printf("Warning: High frequency diffusion<0\n");
            *p->hdif = 0.;
        }
        if (*p->time < 0)       {
            printf("Negative time?\n");
            *p->time = 0.0;
        }
                        
        for (i = 0 ; i < Combs ; i++)   {
            p->c_gain[i] = exp((double)(LOG001 * (p->c_time[i]/esr) /
                                        (gc_gain[i] * *p->time)));
            p->g[i] = *p->hdif;
            p->c_gain[i] = p->c_gain[i] * (1 - p->g[i]);
            p->z[i] = 0.0;
        }

        for (i = 0 ; i < Alpas ; i++)
            p->a_gain[i] = exp((double)(LOG001 * (p->a_time[i]/esr) /
                                        (ga_gain[i] * *p->time)));

        p->prev_time = *p->time;
        p->prev_hdif = *p->hdif;
    }

    for (i = 0 ; i < Combs ; i++)       {

        buf = p->cbuf_cur[i];
        end = (float *)p->caux[i].endp;
        gain = p->c_gain[i];
        in = p->in;
        out = p->out;
        n = ksmps;
        do {
            *out++ += *buf;
            *buf += p->z[i] * p->g[i];
            p->z[i] = *buf;
            *buf *= gain;       
            *buf += *in++;                      
            if (++buf >= end)
                buf = (float *)p->caux[i].auxp;
        } while (--n);
        p->cbuf_cur[i] = buf;
    }

    for (i = 0 ; i < Alpas ; i++)       {

        in = (float *)p->temp.auxp;
        out = p->out;
        n = ksmps;
        do      *in++ = *out++;
        while (--n);

        buf = p->abuf_cur[i];
        end = (float *)p->aaux[i].endp;
        gain = p->a_gain[i];
        in = (float *)p->temp.auxp;
        out = p->out;
        n = ksmps;
        do {
            z = *buf;
            *buf = gain * z + *in++;
            *out++ = z - gain * *buf;
            if (++buf >= end)
                buf = (float *)p->aaux[i].auxp;
        } while (--n);
        p->abuf_cur[i] = buf;
    }
}
